import numpy as np

def LH_Client(input_data, , epsilon, optimal=True):
    
    
    
    
    
    return


def LH_Aggregator(reports, , epsilon, optimal=True):
    
    
    
    
    
    
    return
    
